@@ -1,7 +1,8 @@
-"""A PyConfs Configuration object
+"""A PyConfs Configuration
 
-The `Configuration` is a collection of `ConfigurationEntry` objects. Entries can
-be grouped inside nested Configurations as well.
+The `Configuration` is a dictionary structure containing configuration entries,
+possibly nested inside sections. Each section is its own `Configuration` as
+well.
 """
 
 # Standard library imports
@@ -11,7 +12,7 @@ import pathlib
 import re
 import textwrap
 import warnings
-from collections import UserDict, UserList
+from collections import UserDict, UserList, UserString
 from datetime import date, datetime
 from typing import (
     Any,
@@ -29,16 +30,19 @@ from typing import (
 # PyConfs imports
 from pyconfs import _converters, _exceptions, readers, writers
 
-# Marker for missing keys
-MISSING = object()
-
 
 def _dispatch_to(converter):
-    """Decorator for dispatching methods to converter functions"""
+    """Decorator for dispatching methods to converter functions
+
+    This outer closure stores the converter parameter.
+    """
 
     def _decorator_dispatch_to(func):
+        """Decorate the given function"""
+
         @functools.wraps(func)
         def _wrapper_dispatch_to(self, key: str, **options):
+            """Call the converter function"""
             return converter(value=func(self, key), **options)
 
         return _wrapper_dispatch_to
@@ -51,6 +55,14 @@ class IsConfiguration:
 
 
 class Configuration(UserDict, IsConfiguration):
+    """Consistent handling of configuration formats
+
+    Credit: Originally written for `midgard.config.Configuration`.
+            See https://github.com/kartverket/midgard
+    """
+
+    __slots__ = ("data", "name", "vars", "_source")
+
     def __init__(
         self, name: Optional[str] = None, _vars: Optional[Dict[str, str]] = None
     ) -> None:
@@ -120,6 +132,11 @@ class Configuration(UserDict, IsConfiguration):
         for key, value in entries.items():
             self.update_entry(key=key, value=value, source=source)
 
+    def update_from_config(
+        self, key: str, config: IsConfiguration, source: str = ""
+    ) -> None:
+        """Update the configuration from another configuration"""
+
     def update_from_env(
         self,
         env_paths: Dict[str, Sequence[str]],
@@ -189,6 +206,60 @@ class Configuration(UserDict, IsConfiguration):
         entries = readers.from_str(format, string=string)
         self.update_from_dict(entries, source=source)
 
+    def copy_section(
+        self, from_name: Union[str, List[str]], to_name: Union[str, List[str]]
+    ) -> None:
+        """Copy one section into another, nested sections can be specified as lists"""
+
+        # Get section to copy from, allow nested names
+        section = self.get(from_name)
+        if not isinstance(section, IsConfiguration):
+            raise TypeError(f"{_dotted_name(from_name)} is not a section")
+
+        # Get section to copy to, allow nested names
+        destination = self.get(to_name)
+        if not isinstance(destination, IsConfiguration):
+            raise TypeError(f"{_dotted_name(to_name)} is not a section")
+
+        # Copy section
+        destination.update_entry(
+            key=section.name_stem, value=section, source=f"Copy of {section.name}"
+        )
+
+    def copy_default_section(
+        self, default_name: Union[str, List[str]], copy_recursively: bool = False
+    ) -> None:
+        """Copy contents of one section into all other sections"""
+
+        # Get section to copy from, allow nested names
+        default = self.get(default_name)
+        if not isinstance(default, IsConfiguration):
+            raise TypeError(f"{_dotted_name(default_name)} is not a section")
+
+        # Copy section to all other sections
+        for destination in self.sections:
+            if destination.name == default.name:
+                continue
+
+            if copy_recursively:
+                destination._copy_recursively(default)
+            else:
+                destination.update_entry(
+                    key=default.name_stem,
+                    value=default,
+                    source=f"Copy of {default.name}",
+                )
+
+    def _copy_recursively(self, section: IsConfiguration) -> None:
+        """Copy the given section to all subsections of Configuration recursively"""
+        for destination in self.sections:
+            if destination.name == section.name:
+                continue
+            destination._copy_recursively(section)
+        self.update_entry(
+            key=section.name_stem, value=section, source=f"Copy of {section.name}"
+        )
+
     def get(self, key, default=None):
         """Get a value from the configuration, allow nested keys"""
         if isinstance(key, list):
@@ -204,6 +275,14 @@ class Configuration(UserDict, IsConfiguration):
         else:
             return super().get(key, default)
 
+    @property
+    def name_stem(self) -> str:
+        """Part of name after last dot"""
+        if self.name is None:
+            return ""
+
+        return self.name.rpartition(".")[-1]
+
     @property
     def section_items(self) -> List[Tuple[str, "Configuration"]]:
         """List of section names and sections in Configuration
@@ -328,23 +407,19 @@ class Configuration(UserDict, IsConfiguration):
         **replace_vars: str,
     ) -> Any:
         """Replace values in an entry based on {} format strings"""
-        all_vars = {**self.vars, **replace_vars}
-        value = self.get(key, MISSING)
-        if value is MISSING:
-            raise KeyError(key)
+        all_vars = Variables(**self.vars, **replace_vars, default=default)
+        value = textwrap.dedent(self.data[key]) if dedent else self.data[key]
 
         # Replace variables
-        value = textwrap.dedent(value) if dedent else value
-        replaced = _replace(value, replace_vars=all_vars, default=default)
+        replaced = value.format_map(all_vars)
 
         # Optionally convert value to a different data type
-        if converter is not None:
-            if callable(converter):
-                replaced = converter(replaced)
-            else:
-                replaced = _converters.convert(f"to_{converter}", value=replaced)
-
-        return replaced
+        if converter is None:
+            return replaced
+        elif callable(converter):
+            return converter(replaced)
+        else:
+            return _converters.convert(f"to_{converter}", value=replaced)
 
     def as_dict(self, **other_fields: Any) -> Dict[str, Any]:
         """Convert Configuration to a nested dictionary"""
@@ -524,6 +599,7 @@ class Configuration(UserDict, IsConfiguration):
         return list(super().__dir__()) + list(self.data.keys())
 
     def __getattr__(self, key: str) -> Union["Configuration", Any]:
+        """Get sections and keys using attribute (dot) syntax"""
         try:
             return self[key]
         except KeyError:
@@ -531,6 +607,18 @@ class Configuration(UserDict, IsConfiguration):
                 f"Configuration {self.name} has no entry {key!r}"
             ) from None
 
+    def __setattr__(self, key: str, value: Any) -> None:
+        """Update an entry using assignment syntax on attributes"""
+        if key in self.__slots__:
+            # Update regular attributes as normal
+            super().__setattr__(key, value)
+        else:
+            self.update_entry(key=key, value=value)
+
+    def __setitem__(self, key: str, value: Any) -> None:
+        """Update an entry using assignment syntax on items"""
+        self.update_entry(key=key, value=value)
+
     def __repr__(self):
         """Simple representation of a Configuration"""
         if self.name is None:
@@ -589,6 +677,24 @@ class ConfigurationList(UserList, IsConfiguration):
         """Return configuration as a list of Configurations"""
         return self.data
 
+    def get(self, key, default=None):
+        """Get an entry from the ConfigurationList, allow nested keys"""
+        if isinstance(key, list):
+            first_key, *tail_keys = key
+            value = self.get(first_key, default)
+            if tail_keys:
+                try:
+                    return value.get(list(tail_keys), default)
+                except AttributeError:
+                    return default
+            else:
+                return value
+        else:
+            try:
+                return self[key]
+            except (IndexError, AttributeError):
+                return default
+
     @property
     def leafs(self) -> List[Tuple["Configuration", str, Any]]:
         """List of all keys and values, recursively including subsections"""
@@ -656,48 +762,51 @@ class ConfigurationList(UserList, IsConfiguration):
                 ) from None
 
 
-def _replace(
-    string: str, replace_vars: Dict[str, str], default: Optional[str] = None
-) -> str:
-    """Replace format style variables in a string
+class Variables(UserDict):
+    """Dictionary that allows more flexible format style replacements
 
-    Handles nested replacements by first replacing the replace_vars. Format
-    specifiers (after colon, :) are allowed, but can not contain nested format
-    strings.
+    Instead of using str.format(**dict), you can use str.format_map(Variables)
 
-    This function is used instead of str.format for three reasons. It handles:
+    This will handle:
     - that not all pairs of {...} are replaced at once
     - optional default values for variables that are not specified
     - nested replacements where values of replace_vars may be replaced
       themselves
-
-    Credit: Originally written for `midgard.config.Configuration`.
-            See https://github.com/kartverket/midgard
-
-    Args:
-        string:        Original string
-        replace_vars:  Variables that can be replaced
-        default:       Optional default value used for vars not in replace_vars.
     """
-    matches = re.finditer(r"\{(\w+)(:[^\{\}]*)?\}", string)
-    for match in matches:
-        var = match.group(1)
-        var_expr = match.string[slice(*match.span())]
-        replacement = replace_vars.get(var)
-        if replacement is None:
-            # Default replacements
-            if default is None:
-                continue
-            replacement = default
+
+    def __init__(self, default=None, **vars):
+        """Store optional default value"""
+        super().__init__(**vars)
+        self.default = default
+
+    def __getitem__(self, key):
+        """Handle nested replacements"""
+        value = super().__getitem__(key)
+
+        # Nested replacement for strings
+        if isinstance(value, str):
+            return value.format_map(self)
         else:
-            # Nested replacements
-            if isinstance(replacement, str):
-                replacement = _replace(replacement, replace_vars, default)
+            return value
 
-        # Use str.format to handle format specifiers
-        string = string.replace(var_expr, var_expr.format(**{var: replacement}))
+    def __missing__(self, key):
+        """Handle missing variables"""
+        return UnsetVariable(key) if self.default is None else self.default
 
-    return string
+
+class UnsetVariable(UserString):
+    """Variable that does not have a value yet, used by Variables"""
+
+    def __format__(self, fmt):
+        """Keep the formatting string"""
+        if fmt:
+            return f"{{{self.data}:{fmt}}}"
+        else:
+            return str(self)
+
+    def __str__(self):
+        """Represent unset variables by adding curly braces that can be parsed later"""
+        return f"{{{self.data}}}"
 
 
 def _repr_toml(value: Any) -> str:
@@ -729,3 +838,11 @@ def _repr_toml(value: Any) -> str:
 def _is_nested(sequence) -> bool:
     """Check if the sequence contains nested lists or dictionaries"""
     return any(isinstance(s, (dict, list, IsConfiguration)) for s in sequence)
+
+
+def _dotted_name(nested_name) -> str:
+    """Convert a nested name (with list) to dotted name"""
+    if isinstance(nested_name, list):
+        return ".".join(str(name) for name in nested_name)
+    else:
+        return nested_name
